@@ -5,7 +5,7 @@ use warnings;
 package Hash::Ordered;
 # ABSTRACT: A fast, pure-Perl ordered hash class
 
-our $VERSION = '0.005'; # TRIAL
+our $VERSION = '0.011'; # TRIAL
 
 use Carp ();
 
@@ -57,22 +57,26 @@ use overload
 #pod
 #pod Constructs an object, with an optional list of key-value pairs.
 #pod
+#pod The position of a key corresponds to the first occurrence in the list, but
+#pod the value will be updated if the key is seen more than once.
+#pod
+#pod Current API available since 0.009.
+#pod
 #pod =cut
 
 sub new {
-    my ( $class, @pairs ) = @_;
+    my $class = shift;
 
-    return bless [ {}, [], undef, 0, 0 ], $class unless @pairs;
+    Carp::croak("new() requires key-value pairs") unless @_ % 2 == 0;
 
-    Carp::croak("new() requires key-value pairs") unless @pairs % 2 == 0;
-
-    # must stringify keys for _KEYS array
-    my $self = [
-        {@pairs}, [ map { $_ % 2 == 0 ? ("$pairs[$_]") : () } 0 .. $#pairs ],
-        undef, 0, 0
-    ];
-
-    return bless $self, $class;
+    my ( %data, @keys, $k );
+    while (@_) {
+        # must stringify keys for _KEYS array
+        $k = shift;
+        push @keys, "$k" unless exists $data{$k};
+        $data{$k} = shift;
+    }
+    return bless [ \%data, \@keys, undef, 0, 0 ], $class;
 }
 
 #pod =method clone
@@ -89,12 +93,11 @@ sub new {
 
 sub clone {
     my $self = CORE::shift;
-    my @keys = map { "$_" } @_; # stringify keys
     my $clone;
-    if (@keys) {
+    if (@_) {
         my %subhash;
-        @subhash{@keys} = @{ $self->[_DATA] }{@keys};
-        $clone = [ \%subhash, \@keys, undef, 0, 0 ];
+        @subhash{@_} = @{ $self->[_DATA] }{@_};
+        $clone = [ \%subhash, [ map "$_", @_ ], undef, 0, 0 ];
     }
     elsif ( $self->[_INDX] ) {
         $clone =
@@ -114,7 +117,7 @@ sub clone {
 #pod     $size = $oh->keys;
 #pod
 #pod In list context, returns the ordered list of keys.  In scalar context, returns
-#pod the number of keys.
+#pod the number of elements.
 #pod
 #pod Current API available since 0.005.
 #pod
@@ -137,12 +140,18 @@ sub keys {
 #pod values in order corresponding to those keys.  If a key does not exist, C<undef>
 #pod will be returned for that value.
 #pod
+#pod In scalar context, returns the number of elements.
+#pod
+#pod Current API available since 0.006.
+#pod
 #pod =cut
 
 sub values {
-    my ( $self, @keys ) = @_;
+    my $self = CORE::shift;
     return
-      map { $self->[_DATA]{$_} } ( @keys ? @keys : grep !ref($_), @{ $self->[_KEYS] } );
+      wantarray
+      ? ( map { $self->[_DATA]{$_} } ( @_ ? @_ : grep !ref($_), @{ $self->[_KEYS] } ) )
+      : @{ $self->[_KEYS] } - $self->[_GCNT];
 }
 
 #pod =method get
@@ -170,8 +179,11 @@ sub get {
 sub set {
     my ( $self, $key ) = @_; # don't copy $_[2] in case it's large
     if ( !exists $self->[_DATA]{$key} ) {
+        my $keys = $self->[_KEYS];
+        if ( my $indx = $self->[_INDX] ) {
+            $indx->{$key} = @$keys ? $indx->{ $keys->[-1] } + 1 : 0;
+        }
         CORE::push @{ $self->[_KEYS] }, "$key"; # stringify key
-        $self->[_INDX]{$key} = $#{ $self->[_KEYS] } if $self->[_INDX];
     }
     return $self->[_DATA]{$key} = $_[2];
 }
@@ -257,7 +269,7 @@ sub delete {
 #pod Removes all key-value pairs from the hash.  Returns undef in scalar context
 #pod or an empty list in list context.
 #pod
-#pod Added in version 0.003.
+#pod Current API available since 0.003.
 #pod
 #pod =cut
 
@@ -280,14 +292,16 @@ sub clear {
 #pod =cut
 
 sub push {
-    my ( $self, @pairs ) = @_;
-    my ( $data, $keys, $indx ) = @$self;
-    while (@pairs) {
-        my ( $k, $v ) = splice( @pairs, 0, 2 );
+    my $self = CORE::shift;
+    my ( $data, $keys ) = @$self;
+    while (@_) {
+        my ( $k, $v ) = splice( @_, 0, 2 );
         $self->delete($k) if exists $data->{$k};
         $data->{$k} = $v;
+        if ( my $indx = $self->[_INDX] ) {
+            $indx->{$k} = @$keys ? $indx->{ $keys->[-1] } + 1 : 0;
+        }
         CORE::push @$keys, "$k"; # stringify keys
-        $indx->{$k} = $#$keys if $indx;
     }
     return @$keys - $self->[_GCNT];
 }
@@ -295,8 +309,11 @@ sub push {
 #pod =method pop
 #pod
 #pod     ($key, $value) = $oh->pop;
+#pod     $value = $oh->pop;
 #pod
 #pod Removes and returns the last key-value pair in the ordered hash.
+#pod In scalar context, only the value is returned.  If the hash is empty,
+#pod the returned key and value will be C<undef>.
 #pod
 #pod =cut
 
@@ -308,7 +325,7 @@ sub pop {
     }
     else {
         my $key = CORE::pop @{ $self->[_KEYS] };
-        return $key, delete $self->[_DATA]{$key};
+        return defined($key) ? ( $key, delete $self->[_DATA]{$key} ) : ();
     }
 }
 
@@ -325,14 +342,14 @@ sub pop {
 #pod =cut
 
 sub unshift {
-    my ( $self, @pairs ) = @_;
-    my ( $data, $keys, $indx ) = @$self;
-    while (@pairs) {
-        my ( $k, $v ) = splice( @pairs, -2, 2 );
+    my $self = CORE::shift;
+    my ( $data, $keys ) = @$self;
+    while (@_) {
+        my ( $k, $v ) = splice( @_, -2, 2 );
         $self->delete($k) if exists $data->{$k};
         $data->{$k} = $v;
         CORE::unshift @$keys, "$k"; # stringify keys
-        $indx->{$k} = -( ++$self->[_OFFS] ) if $indx;
+        $self->[_INDX]{$k} = -( ++$self->[_OFFS] ) if $self->[_INDX];
     }
     return @$keys - $self->[_GCNT];
 }
@@ -340,8 +357,11 @@ sub unshift {
 #pod =method shift
 #pod
 #pod     ($key, $value) = $oh->shift;
+#pod     $value = $oh->shift;
 #pod
 #pod Removes and returns the first key-value pair in the ordered hash.
+#pod In scalar context, only the value is returned.  If the hash is empty,
+#pod the returned key and value will be C<undef>.
 #pod
 #pod =cut
 
@@ -353,7 +373,7 @@ sub shift {
     }
     else {
         my $key = CORE::shift @{ $self->[_KEYS] };
-        return $key, delete $self->[_DATA]{$key};
+        return defined($key) ? ( $key, delete $self->[_DATA]{$key} ) : ();
     }
 }
 
@@ -368,9 +388,9 @@ sub shift {
 #pod =cut
 
 sub merge {
-    my ( $self, @pairs ) = @_;
-    while (@pairs) {
-        my ( $k, $v ) = splice( @pairs, -2, 2 );
+    my $self = CORE::shift;
+    while (@_) {
+        my ( $k, $v ) = splice( @_, 0, 2 );
         if ( !exists $self->[_DATA]{$k} ) {
             my $size = CORE::push @{ $self->[_KEYS] }, "$k"; # stringify key
             $self->[_INDX]{$k} = $size - 1 if $self->[_INDX];
@@ -393,9 +413,10 @@ sub merge {
 #pod =cut
 
 sub as_list {
-    my ( $self, @keys ) = @_;
-    @keys = grep !ref($_), @{ $self->[_KEYS] } unless @keys;
-    return map { ; $_ => $self->[_DATA]{$_} } @keys;
+    my $self = CORE::shift;
+    return
+      map { ; $_ => $self->[_DATA]{$_} }
+      ( @_ ? @_ : grep !ref($_), @{ $self->[_KEYS] } );
 }
 
 #pod =method iterator
@@ -418,6 +439,7 @@ sub as_list {
 #pod
 #pod =cut
 
+# usually we avoid copying keys in @_; here we must for the closure
 sub iterator {
     my ( $self, @keys ) = @_;
     @keys = grep !ref($_), @{ $self->[_KEYS] } unless @keys;
@@ -622,7 +644,7 @@ sub FIRSTKEY {
 }
 
 sub NEXTKEY {
-    return $_[0]->[_ITER]->();
+    return defined( $_[0]->[_ITER] ) ? $_[0]->[_ITER]->() : undef;
 }
 
 sub SCALAR {
@@ -646,7 +668,7 @@ Hash::Ordered - A fast, pure-Perl ordered hash class
 
 =head1 VERSION
 
-version 0.005
+version 0.011
 
 =head1 SYNOPSIS
 
@@ -707,6 +729,11 @@ L<Hash::Ordered::Benchmarks>.
 
 Constructs an object, with an optional list of key-value pairs.
 
+The position of a key corresponds to the first occurrence in the list, but
+the value will be updated if the key is seen more than once.
+
+Current API available since 0.009.
+
 =head2 clone
 
     $oh2 = $oh->clone;
@@ -723,7 +750,7 @@ in the original will have the value C<undef>.
     $size = $oh->keys;
 
 In list context, returns the ordered list of keys.  In scalar context, returns
-the number of keys.
+the number of elements.
 
 Current API available since 0.005.
 
@@ -737,6 +764,10 @@ the ordered values of the entire hash.  If a list of keys is given, returns
 values in order corresponding to those keys.  If a key does not exist, C<undef>
 will be returned for that value.
 
+In scalar context, returns the number of elements.
+
+Current API available since 0.006.
+
 =head2 get
 
     $value = $oh->get("some key");
@@ -770,7 +801,7 @@ Removes a key-value pair from the hash and returns the value.
 Removes all key-value pairs from the hash.  Returns undef in scalar context
 or an empty list in list context.
 
-Added in version 0.003.
+Current API available since 0.003.
 
 =head2 push
 
@@ -785,8 +816,11 @@ Returns the number of keys after the push is complete.
 =head2 pop
 
     ($key, $value) = $oh->pop;
+    $value = $oh->pop;
 
 Removes and returns the last key-value pair in the ordered hash.
+In scalar context, only the value is returned.  If the hash is empty,
+the returned key and value will be C<undef>.
 
 =head2 unshift
 
@@ -801,8 +835,11 @@ Returns the number of keys after the unshift is complete.
 =head2 shift
 
     ($key, $value) = $oh->shift;
+    $value = $oh->shift;
 
 Removes and returns the first key-value pair in the ordered hash.
+In scalar context, only the value is returned.  If the hash is empty,
+the returned key and value will be C<undef>.
 
 =head2 merge
 
@@ -972,6 +1009,8 @@ If you want to access the underlying object for method calls, use C<tied>:
 
     tied( %hash )->unshift( @data );
 
+Tied hash API available since 0.005.
+
 =head1 CAVEATS
 
 =head2 Deletion and order modification with push, pop, etc.
@@ -1033,42 +1072,69 @@ benchmarking results, see L<Hash::Ordered::Benchmarks>.
 
 =head2 Tie modules
 
-The following modules offer some sort of tie interface.  I don't like ties, in
-general, because of the extra indirection involved over a direct method call,
-but if you are willing to pay that penalty, you might want to try one of these.
+The following modules offer some sort of tie interface.  I don't like ties,
+in general, because of the extra indirection involved over a direct method
+call. Still, you can make any tied interface into a faster OO one with
+C<tied>:
 
-L<Tie::IxHash> is probably the most well known and includes an OO API.  If its
-warts and performance profile aren't a problem, it might serve.
+    tied( %tied_hash )->FETCH($key);
 
-L<Tie::Hash::Indexed> is implemented in XS and thus seems promising if pure-Perl
-isn't a criterion; it often fails tests on Perl 5.18 and above due to the hash
-randomization change.
+L<Tie::Hash::Indexed> is implemented in XS and thus seems promising if
+pure-Perl isn't a criterion; it generally fails tests on Perl 5.18 and
+above due to the hash randomization change.  Despite being XS, it is slower
+than Hash::Ordered at everything exception creation and deletion.
 
-These other modules have very specific designs/limitations and I didn't find
-any of them suitable for general purpose use:
+L<Tie::IxHash> is probably the most well known and includes an OO API.
+Given the performance problems it has, "well known" is the only real reason
+to use it.
 
-* L<Tie::Array::AsHash> — array elements split with separator; tie API only
-* L<Tie::Hash::Array> — ordered alphabetically; tie API only
-* L<Tie::InsertOrderHash> — ordered by insertion; tie API only
-* L<Tie::LLHash> — linked-list implementation; quite slow
-* L<Tie::StoredOrderHash> — ordered by last update; tie API only
+These other modules below have very specific designs/limitations and I
+didn't find any of them suitable for general purpose use:
+
+=over 4
+
+=item *
+
+L<Tie::Array::AsHash> — array elements split with separator; tie API only
+
+=item *
+
+L<Tie::Hash::Array> — ordered alphabetically; tie API only
+
+=item *
+
+L<Tie::InsertOrderHash> — ordered by insertion; tie API only
+
+=item *
+
+L<Tie::LLHash> — linked-list implementation; quite slow
+
+=item *
+
+L<Tie::StoredOrderHash> — ordered by last update; tie API only
+
+=back
 
 =head2 Other ordered hash modules
 
 Other modules stick with an object-oriented API, with a wide variety of
 implementation approaches.
 
-L<Array::AsHash> is essentially an inverse implementation from Hash::Ordered.
-It keeps pairs in an array and uses a hash to index into the array.  I think
-this indirection makes hash-like operations slower, but getting the list of
-pairs back out is much faster.  It takes an arrayref to initialize, but can
-shallow copy it if needed.  I think this is a reasonable alternative if static
-construction and listing out contents is more common than individual item
-access.
+L<Array::AsHash> is essentially an inverse implementation from
+Hash::Ordered.  It keeps pairs in an array and uses a hash to index into
+the array.  This indirection would already make hash-like operations
+slower, but the specific implementation makes it even worse, with
+abstractions and function calls that make getting or setting individual
+items up to 10x slower than Hash::Ordered.
 
-These other modules have restrictions or particularly complicated
-implementations (often relying on C<tie>) and thus I didn't think any of them
-really suitable for use:
+However, C<Array::AsHash> takes an arrayref to initialize, which is very
+fast and can return the list of pairs faster, too.  If you mostly create
+and list out very large ordered hashes and very rarely touch individual
+entries, I think this could be something to very cautiously consider.
+
+These other modules below have restrictions or particularly complicated
+implementations (often relying on C<tie>) and thus I didn't think any of
+them really suitable for use:
 
 =over 4
 
